@@ -4,9 +4,9 @@ import os
 import argparse
 from multiprocessing import freeze_support
 from .langserver import LangServer
-from .jsonrpc import JSONRPC2Connection, ReadWriter
+from .jsonrpc import JSONRPC2Connection, ReadWriter, path_from_uri
 from .parse_fortran import process_file, detect_fixed_format
-__version__ = '0.6.2'
+__version__ = '1.0.1'
 
 
 def error_exit(error_str):
@@ -28,7 +28,7 @@ def main():
         help="Do not include type members in document symbol results"
     )
     parser.add_argument(
-        '--incrmental_sync', action="store_true",
+        '--incremental_sync', '--incrmental_sync', action="store_true",
         help="Use incremental document synchronization (beta)"
     )
     parser.add_argument(
@@ -39,23 +39,51 @@ def main():
         '--lowercase_intrinsics', action="store_true",
         help="Use lowercase for intrinsics and keywords in autocomplete requests"
     )
+    parser.add_argument(
+        '--use_signature_help', action="store_true",
+        help="Use signature help instead of subroutine/function snippets"
+    )
+    parser.add_argument(
+        '--variable_hover', action="store_true",
+        help="Show hover information for variables (default: subroutines/functions only)"
+    )
+    parser.add_argument(
+        '--debug_log', action="store_true",
+        help="Generate debug log in project root folder"
+    )
     group = parser.add_argument_group("DEBUG", "Options for debugging language server")
     group.add_argument(
         '--debug_parser', action="store_true",
         help="Test source code parser on specified file"
     )
+    group.add_argument(
+        '--debug_diagnostics', action="store_true",
+        help="Test diagnostic notifications for specified file"
+    )
     group.add_argument(
         '--debug_symbols', action="store_true",
         help="Test symbol request for specified file"
     )
+    group.add_argument(
+        '--debug_workspace_symbols', type=str,
+        help="Test workspace/symbol request"
+    )
     group.add_argument(
         '--debug_completion', action="store_true",
         help="Test completion request for specified file and position"
     )
+    group.add_argument(
+        '--debug_signature', action="store_true",
+        help="Test signatureHelp request for specified file and position"
+    )
     group.add_argument(
         '--debug_definition', action="store_true",
         help="Test definition request for specified file and position"
     )
+    group.add_argument(
+        '--debug_hover', action="store_true",
+        help="Test hover request for specified file and position"
+    )
     group.add_argument(
         '--debug_filepath', type=str,
         help="File path for language server tests"
@@ -76,14 +104,19 @@ def main():
     if args.version:
         print("{0}".format(__version__))
         sys.exit(0)
-    debug_server = (args.debug_symbols or args.debug_completion
-                    or (args.debug_rootpath is not None))
+    debug_server = (args.debug_diagnostics or args.debug_symbols
+                    or args.debug_completion or args.debug_signature
+                    or args.debug_definition or args.debug_hover
+                    or (args.debug_rootpath is not None)
+                    or (args.debug_workspace_symbols is not None))
     #
     settings = {
         "symbol_include_mem": (not args.symbol_skip_mem),
-        "sync_type": 2 if args.incrmental_sync else 1,
+        "sync_type": 2 if args.incremental_sync else 1,
         "autocomplete_no_prefix": args.autocomplete_no_prefix,
-        "lowercase_intrinsics": args.lowercase_intrinsics
+        "lowercase_intrinsics": args.lowercase_intrinsics,
+        "use_signature_help": args.use_signature_help,
+        "variable_hover": args.variable_hover
     }
     #
     if args.debug_parser:
@@ -115,7 +148,7 @@ def main():
         tmpin = os.fdopen(prb, 'rb')
         tmpout = os.fdopen(pwb, 'wb')
         s = LangServer(conn=JSONRPC2Connection(ReadWriter(tmpin, tmpout)),
-                       logLevel=0, settings=settings)
+                       debug_log=args.debug_log, settings=settings)
         #
         if args.debug_rootpath:
             dir_exists = os.path.isdir(args.debug_rootpath)
@@ -124,9 +157,7 @@ def main():
             print('\nTesting "initialize" request:')
             print('  Root = "{0}"\n'.format(args.debug_rootpath))
             s.serve_initialize({
-                "params": {
-                    "rootPath": args.debug_rootpath
-                }
+                "params": {"rootPath": args.debug_rootpath}
             })
             if len(s.post_messages) == 0:
                 print("  Succesful")
@@ -139,14 +170,31 @@ def main():
             for mod_dir in s.mod_dirs:
                 print("    {0}".format(mod_dir))
         #
+        if args.debug_diagnostics:
+            print('\nTesting "textDocument/publishDiagnostics" notification:')
+            check_request_params(args, loc_needed=False)
+            s.serve_onSave({
+                "params": {
+                    "textDocument": {"uri": args.debug_filepath}
+                }
+            })
+            diag_results, diag_exp = s.get_diagnostics(args.debug_filepath)
+            if diag_results is not None:
+                sev_map = ["", "ERROR", "WARNING"]
+                if len(diag_results) == 0:
+                    print("\nNo errors or warnings")
+                else:
+                    print("\nReported errors or warnings:")
+                for diag in diag_results:
+                    sline = diag["range"]["start"]["line"]
+                    message = diag["message"]
+                    sev = sev_map[diag["severity"]]
+                    print('  {0:5d}:{1}  "{2}"'.format(sline,
+                          sev, message))
+        #
         if args.debug_symbols:
             print('\nTesting "textDocument/documentSymbol" request:')
-            if args.debug_filepath is None:
-                error_exit("'debug_filepath' not specified for document symbol test")
-            file_exists = os.path.isfile(args.debug_filepath)
-            if file_exists is False:
-                error_exit("Specified 'debug_filepath' does not exist")
-            print('  File = "{0}"\n'.format(args.debug_filepath))
+            check_request_params(args, loc_needed=False)
             s.serve_onSave({
                 "params": {
                     "textDocument": {"uri": args.debug_filepath}
@@ -166,20 +214,28 @@ def main():
                 print('  line {2:5d}  symbol -> {1:3d}:{0:30} parent = {3}'.format(symbol["name"],
                       symbol["kind"], sline, parent))
         #
+        if args.debug_workspace_symbols is not None:
+            print('\nTesting "workspace/symbol" request:')
+            if args.debug_rootpath is None:
+                error_exit("'debug_rootpath' not specified for debug request")
+            symbol_results = s.serve_workspace_symbol({
+                "params": {
+                    "query": args.debug_workspace_symbols
+                }
+            })
+            for symbol in symbol_results:
+                path = path_from_uri(symbol["location"]["uri"])
+                sline = symbol["location"]["range"]["start"]["line"]
+                if "containerName" in symbol:
+                    parent = symbol["containerName"]
+                else:
+                    parent = "null"
+                print('  {2}::{3:d}  symbol -> {1:3d}:{0:30} parent = {4}'.format(symbol["name"],
+                      symbol["kind"], os.path.relpath(path, args.debug_rootpath), sline, parent))
+        #
         if args.debug_completion:
             print('\nTesting "textDocument/completion" request:')
-            if args.debug_filepath is None:
-                error_exit("'debug_filepath' not specified for completion test")
-            file_exists = os.path.isfile(args.debug_filepath)
-            if file_exists is False:
-                error_exit("Specified 'debug_filepath' does not exist")
-            print('  File = "{0}"'.format(args.debug_filepath))
-            if args.debug_line is None:
-                error_exit("'debug_line' not specified for completion test")
-            print('  Line = {0}'.format(args.debug_line))
-            if args.debug_char is None:
-                error_exit("'debug_char' not specified for completion test")
-            print('  Char = {0}\n'.format(args.debug_char))
+            check_request_params(args)
             s.serve_onSave({
                 "params": {
                     "textDocument": {"uri": args.debug_filepath}
@@ -193,50 +249,116 @@ def main():
             })
             print('  Results:')
             for obj in completion_results['items']:
-                print('    {0}: {1}'.format(obj['kind'], obj['label']))
+                print('    {0}: {1} -> {2}'.format(obj['kind'], obj['label'], obj['detail']))
         #
-        if args.debug_definition:
-            print('\nTesting "textDocument/definition" request:')
-            if args.debug_filepath is None:
-                error_exit("'debug_filepath' not specified for definition test")
-            file_exists = os.path.isfile(args.debug_filepath)
-            if file_exists is False:
-                error_exit("Specified 'debug_filepath' does not exist")
-            print('  File = "{0}"'.format(args.debug_filepath))
-            if args.debug_line is None:
-                error_exit("'debug_line' not specified for definition test")
-            print('  Line = {0}'.format(args.debug_line))
-            if args.debug_char is None:
-                error_exit("'debug_char' not specified for definition test")
-            print('  Char = {0}\n'.format(args.debug_char))
+        if args.debug_signature:
+            print('\nTesting "textDocument/signatureHelp" request:')
+            check_request_params(args)
             s.serve_onSave({
                 "params": {
                     "textDocument": {"uri": args.debug_filepath}
                 }
             })
-            completion_results = s.serve_definition({
+            signature_results = s.serve_signature({
+                "params": {
+                    "textDocument": {"uri": args.debug_filepath},
+                    "position": {"line": args.debug_line-1, "character": args.debug_char-1}
+                }
+            })
+            if len(signature_results['signatures']) == 0:
+                print('  No Results')
+            else:
+                print('  Results:')
+                active_param = signature_results.get('activeParameter', 0)
+                print('    Active param = {0}'.format(active_param))
+                active_signature = signature_results.get('activeSignature', 0)
+                print('    Active sig   = {0}'.format(active_signature))
+                for i, signature in enumerate(signature_results['signatures']):
+                    print('    {0}'.format(signature['label']))
+                    for j, obj in enumerate(signature['parameters']):
+                        if (i == active_signature) and (j == active_param):
+                            active_mark = '*'
+                        else:
+                            active_mark = ' '
+                        arg_desc = obj.get('documentation')
+                        if arg_desc is not None:
+                            print('{2}     {0} :: {1}'.format(arg_desc, obj['label'], active_mark))
+                        else:
+                            print('{1}     {0}'.format(obj['label'], active_mark))
+        #
+        if args.debug_definition:
+            print('\nTesting "textDocument/definition" request:')
+            check_request_params(args)
+            s.serve_onSave({
+                "params": {
+                    "textDocument": {"uri": args.debug_filepath}
+                }
+            })
+            definition_results = s.serve_definition({
                 "params": {
                     "textDocument": {"uri": args.debug_filepath},
                     "position": {"line": args.debug_line-1, "character": args.debug_char-1}
                 }
             })
             print('  Result:')
-            if completion_results is None:
+            if definition_results is None:
                 print('    No result found!')
             else:
-                print('    URI  = "{0}"'.format(completion_results['uri']))
-                print('    Line = {0}'.format(completion_results['range']['start']['line']+1))
-                print('    Char = {0}'.format(completion_results['range']['start']['character']+1))
+                print('    URI  = "{0}"'.format(definition_results['uri']))
+                print('    Line = {0}'.format(definition_results['range']['start']['line']+1))
+                print('    Char = {0}'.format(definition_results['range']['start']['character']+1))
+        #
+        if args.debug_hover:
+            print('\nTesting "textDocument/hover" request:')
+            check_request_params(args)
+            s.serve_onSave({
+                "params": {
+                    "textDocument": {"uri": args.debug_filepath}
+                }
+            })
+            hover_results = s.serve_hover({
+                "params": {
+                    "textDocument": {"uri": args.debug_filepath},
+                    "position": {"line": args.debug_line-1, "character": args.debug_char-1}
+                }
+            })
+            print('  Result:')
+            if hover_results is None:
+                print('    No result found!')
+            else:
+                contents = hover_results['contents']
+                print('=======')
+                if isinstance(contents, dict):
+                    print(contents['value'])
+                else:
+                    print(contents)
+                print('=======')
         tmpout.close()
         tmpin.close()
     #
     else:
         stdin, stdout = _binary_stdio()
         s = LangServer(conn=JSONRPC2Connection(ReadWriter(stdin, stdout)),
-                       logLevel=0, settings=settings)
+                       debug_log=args.debug_log, settings=settings)
         s.run()
 
 
+def check_request_params(args, loc_needed=True):
+    if args.debug_filepath is None:
+        error_exit("'debug_filepath' not specified for debug request")
+    file_exists = os.path.isfile(args.debug_filepath)
+    if file_exists is False:
+        error_exit("Specified 'debug_filepath' does not exist")
+    print('  File = "{0}"'.format(args.debug_filepath))
+    if loc_needed:
+        if args.debug_line is None:
+            error_exit("'debug_line' not specified for debug request")
+        print('  Line = {0}'.format(args.debug_line))
+        if args.debug_char is None:
+            error_exit("'debug_char' not specified for debug request")
+        print('  Char = {0}\n'.format(args.debug_char))
+
+
 def print_children(obj, indent=""):
     for child in obj.get_children():
         print("  {0}{1}: {2}".format(indent, child.get_type(), child.FQSN))
