@@ -4,17 +4,20 @@ import os
 import traceback
 import re
 # Local modules
+from fortls.jsonrpc import path_to_uri, path_from_uri
 from fortls.parse_fortran import process_file, read_use_stmt, read_var_def, \
     detect_fixed_format, detect_comment_start
 from fortls.objects import find_in_scope, get_use_tree
-from fortls.intrinsics import get_keywords, get_intrinsics, get_intrinsic_modules, \
-    set_lowercase_intrinsics
+from fortls.intrinsics import get_keywords, load_intrinsics, set_lowercase_intrinsics
 
 log = logging.getLogger(__name__)
 PY3K = sys.version_info >= (3, 0)
+if not PY3K:
+    import io
 # Global regexes
 FORTRAN_EXT_REGEX = re.compile(r'^\.F(77|90|95|03|08|OR|PP)?$', re.I)
 OBJBREAK_REGEX = re.compile(r'[\/\-(.,+*<>=$: ]', re.I)
+INT_STMNT_REGEX = re.compile(r'^[ ]*[a-z]*$', re.I)
 WORD_REGEX = re.compile(r'[a-z_][a-z0-9_]*', re.I)
 CALL_REGEX = re.compile(r'[ ]*CALL[ ]+[a-z0-9_%]*$', re.I)
 TYPE_STMNT_REGEX = re.compile(r'[ ]*(TYPE|CLASS)[ ]*(IS)?[ ]*$', re.I)
@@ -28,38 +31,19 @@ FIXED_CONT_REGEX = re.compile(r'(     [\S])')
 FREE_OPT_CONT_REGEX = re.compile(r'([ ]*&)')
 
 
-def path_from_uri(uri):
-    # Convert file uri to path (strip html like head part)
-    if not uri.startswith("file://"):
-        return uri
-    if os.name == "nt":
-        _, path = uri.split("file:///", 1)
-        path = path.replace("%3A", ":")  # Replace escaped colon in Windows paths
-    else:
-        _, path = uri.split("file://", 1)
-    return path
-
-
-def path_to_uri(path):
-    # Convert path to file uri (add html like head part)
-    if os.name == "nt":
-        return "file:///"+path
-    else:
-        return "file://"+path
-
-
 def read_file_split(filepath):
     # Read and add file from disk
     try:
         if PY3K:
-            with open(filepath, 'r', encoding="utf-8") as fhandle:
+            with open(filepath, 'r', encoding='utf-8', errors='replace') as fhandle:
                 contents = re.sub(r'\t', r'  ', fhandle.read())
                 contents_split = contents.splitlines()
         else:
-            with open(filepath, 'r') as fhandle:
+            with io.open(filepath, 'r', encoding='utf-8', errors='replace') as fhandle:
                 contents = re.sub(r'\t', r'  ', fhandle.read())
                 contents_split = contents.splitlines()
     except:
+        log.error("Could not read/decode file %s", filepath, exc_info=True)
         return None, 'Could not read/decode file'
     else:
         return contents_split, None
@@ -75,6 +59,7 @@ def init_file(filepath):
         fixed_flag = detect_fixed_format(contents_split)
         ast_new = process_file(contents_split, True, filepath, fixed_flag)
     except:
+        log.error("Error while parsing file %s", filepath, exc_info=True)
         return None, 'Error during parsing'
     # Construct new file object and add to workspace
     tmp_obj = {
@@ -95,15 +80,15 @@ def tokenize_line(line):
             if char == string_char:
                 in_string = False
             continue
-        if char == '(':
+        if (char == '(') or (char == '['):
             paren_list[level].append([i, len(line)])
             level += 1
             if len(paren_list) < level+1:
                 paren_list.append([])
-        elif char == ')':
+        elif (char == ')') or (char == ']'):
             paren_list[level-1][-1][1] = i
             level -= 1
-        elif char == "'" or char == '"':
+        elif (char == "'") or (char == '"'):
             in_string = True
             string_char = char
     return paren_split(line, paren_list[:-1])
@@ -121,7 +106,7 @@ def paren_split(line, paren_list):
             ranges = []
             if len(paren_list) > ilev+1:
                 for lower_group in paren_list[ilev+1]:
-                    if lower_group[0] > i1 and lower_group[1] <= i2:
+                    if (lower_group[0] > i1) and (lower_group[1] <= i2):
                         tmp_str += line[i3:lower_group[0]]
                         ranges.append([i3, lower_group[0]])
                         i3 = lower_group[1] + 1
@@ -216,6 +201,11 @@ def climb_type_tree(var_stack, curr_scope, obj_tree):
         # Find next variable by name in scope
         var_name = var_stack[iVar].strip().lower()
         var_obj, new_scope = find_in_scope(type_scope, var_name, obj_tree)
+        # Set scope to declaration location if variable is inherited
+        if new_scope.get_type() == 4:
+            for in_child in new_scope.in_children:
+                if (in_child.name.lower() == var_name) and (in_child.parent is not None):
+                    curr_scope = in_child.parent
         if var_obj is not None:
             type_name = get_type_name(var_obj)
         else:
@@ -270,14 +260,16 @@ def get_line(line, character, file_obj):
 
 def apply_change(contents_split, change):
     """Apply a change to the document."""
-    text = change['text']
+    text = change.get('text', "")
     change_range = change.get('range')
+    if not PY3K:
+        text = text.encode('utf-8')
     if len(text) == 0:
         text_split = [""]
     else:
         text_split = text.splitlines()
         # Check for ending newline
-        if text[-1] == "\n" or text[-1] == "\r":
+        if (text[-1] == "\n") or (text[-1] == "\r"):
             text_split.append("")
 
     if change_range is None:
@@ -294,7 +286,7 @@ def apply_change(contents_split, change):
         return contents_split + text_split, -1
 
     # Check for single line edit
-    if start_line == end_line and len(text_split) == 1:
+    if (start_line == end_line) and (len(text_split) == 1):
         prev_line = contents_split[start_line]
         contents_split[start_line] = prev_line[:start_col] + text + prev_line[end_col:]
         return contents_split, start_line
@@ -302,14 +294,14 @@ def apply_change(contents_split, change):
     # Apply standard change to document
     new_contents = []
     for i, line in enumerate(contents_split):
-        if i < start_line or i > end_line:
+        if (i < start_line) or (i > end_line):
             new_contents.append(line)
             continue
 
         if i == start_line:
             for j, change_line in enumerate(text_split):
                 if j == 0:
-                    new_contents.append(line[:start_col]+change_line)
+                    new_contents.append(line[:start_col] + change_line)
                 else:
                     new_contents.append(change_line)
 
@@ -319,7 +311,7 @@ def apply_change(contents_split, change):
 
 
 class LangServer:
-    def __init__(self, conn, logLevel=0, settings=None):
+    def __init__(self, conn, debug_log=False, settings={}):
         self.conn = conn
         self.running = True
         self.root_path = None
@@ -329,30 +321,17 @@ class LangServer:
         self.obj_tree = {}
         self.mod_dirs = []
         self.excl_paths = []
+        self.excl_suffixes = []
         self.post_messages = []
         self.streaming = True
-        self.symbol_include_mem = True
-        self.autocomplete_no_prefix = False
-        self.lowercase_intrinsics = False
-        self.sync_type = 1
-        if logLevel == 0:
-            logging.basicConfig(level=logging.ERROR)
-        elif logLevel == 1:
-            logging.basicConfig(level=logging.DEBUG)
-        if settings is not None:
-            if "symbol_include_mem" in settings:
-                self.symbol_include_mem = settings["symbol_include_mem"]
-            if "sync_type" in settings:
-                self.sync_type = settings["sync_type"]
-            if "autocomplete_no_prefix" in settings:
-                self.autocomplete_no_prefix = settings["autocomplete_no_prefix"]
-            if "lowercase_intrinsics" in settings:
-                self.lowercase_intrinsics = settings["lowercase_intrinsics"]
-        if self.lowercase_intrinsics:
-            set_lowercase_intrinsics()
-        self.intrinsics = get_intrinsics()
-        for module in get_intrinsic_modules():
-            self.obj_tree[module.FQSN] = [module, None]
+        self.debug_log = debug_log
+        # Get launch settings
+        self.symbol_include_mem = settings.get("symbol_include_mem", True)
+        self.sync_type = settings.get("sync_type", 1)
+        self.autocomplete_no_prefix = settings.get("autocomplete_no_prefix", False)
+        self.lowercase_intrinsics = settings.get("lowercase_intrinsics", False)
+        self.use_signature_help = settings.get("use_signature_help", False)
+        self.variable_hover = settings.get("variable_hover", False)
 
     def post_message(self, message, type=1):
         self.conn.send_notification("window/showMessage", {
@@ -380,18 +359,22 @@ class LangServer:
         def noop(request):
             return None
         # Request handler
-        log.info("REQUEST %s %s", request.get("id"), request.get("method"))
+        log.debug("REQUEST %s %s", request.get("id"), request.get("method"))
         handler = {
             "initialize": self.serve_initialize,
             "textDocument/documentSymbol": self.serve_document_symbols,
             "textDocument/completion": self.serve_autocomplete,
+            "textDocument/signatureHelp": self.serve_signature,
             "textDocument/definition": self.serve_definition,
             "textDocument/references": self.serve_references,
             "textDocument/hover": self.serve_hover,
             "textDocument/didOpen": self.serve_onSave,
             "textDocument/didSave": self.serve_onSave,
-            "textDocument/didClose": self.serve_onSave,
+            "textDocument/didClose": self.serve_onClose,
             "textDocument/didChange": self.serve_onChange,
+            "initialized": noop,
+            "workspace/didChangeWatchedFiles": noop,
+            "workspace/symbol": self.serve_workspace_symbol,
             "$/cancelRequest": noop,
             "shutdown": noop,
             "exit": self.serve_exit,
@@ -413,8 +396,8 @@ class LangServer:
         except JSONRPC2Error as e:
             self.conn.write_error(
                 request["id"], code=e.code, message=e.message, data=e.data)
+            log.warning("RPC error handling request %s", request, exc_info=True)
         except Exception as e:
-            log.warning("handler for %s failed", request, exc_info=True)
             self.conn.write_error(
                 request["id"],
                 code=-32603,
@@ -436,16 +419,35 @@ class LangServer:
         config_path = os.path.join(self.root_path, ".fortls")
         config_exists = os.path.isfile(config_path)
         if config_exists:
-            import json
-            with open(config_path, 'r') as fhandle:
-                config_dict = json.load(fhandle)
-                if "excl_paths" in config_dict:
-                    for excl_path in config_dict["excl_paths"]:
+            try:
+                import json
+                with open(config_path, 'r') as fhandle:
+                    config_dict = json.load(fhandle)
+                    for excl_path in config_dict.get("excl_paths", []):
                         self.excl_paths.append(os.path.join(self.root_path, excl_path))
-                if "mod_dirs" in config_dict:
-                    for mod_dir in config_dict["mod_dirs"]:
-                        self.mod_dirs.append(os.path.join(self.root_path, mod_dir))
-        if len(self.mod_dirs) == 1:  # Recursively add sub-directories
+                    for mod_dir in config_dict.get("mod_dirs", []):
+                        dir_path = os.path.join(self.root_path, mod_dir)
+                        if os.path.isdir(dir_path):
+                            self.mod_dirs.append(dir_path)
+                    self.excl_suffixes = config_dict.get("excl_suffixes", [])
+                    self.lowercase_intrinsics = config_dict.get("lowercase_intrinsics", self.lowercase_intrinsics)
+                    self.debug_log = config_dict.get("debug_log", self.debug_log)
+            except:
+                self.post_messages.append([1, "Error while parsing '.fortls' settings file"])
+        # Setup logging
+        if self.debug_log and (self.root_path != ""):
+            logging.basicConfig(filename=os.path.join(self.root_path, "fortls_debug.log"),
+                                level=logging.DEBUG, filemode='w')
+            log.debug("REQUEST %s %s", request.get("id"), request.get("method"))
+            self.post_messages.append([3, "FORTLS debugging enabled"])
+        # Load intrinsics
+        if self.lowercase_intrinsics:
+            set_lowercase_intrinsics()
+        self.statements, self.keywords, self.intrinsic_funs, self.intrinsic_mods = load_intrinsics()
+        for module in self.intrinsic_mods:
+            self.obj_tree[module.FQSN] = [module, None]
+        # Recursively add sub-directories
+        if len(self.mod_dirs) == 1:
             self.mod_dirs = []
             for dirName, subdirList, fileList in os.walk(self.root_path):
                 if self.excl_paths.count(dirName) > 0:
@@ -463,23 +465,85 @@ class LangServer:
         # Initialize workspace
         self.workspace_init()
         #
-        return {
-            "capabilities": {
-                "completionProvider": {
-                    "resolveProvider": False,
-                    "triggerCharacters": ["%"]
-                },
-                "definitionProvider": True,
-                "documentSymbolProvider": True,
-                "referencesProvider": True,
-                "hoverProvider": True,
-                "textDocumentSync": self.sync_type
-            }
+        server_capabilities = {
+            "completionProvider": {
+                "resolveProvider": False,
+                "triggerCharacters": ["%"]
+            },
+            "definitionProvider": True,
+            "documentSymbolProvider": True,
+            "referencesProvider": True,
+            "hoverProvider": True,
+            "workspaceSymbolProvider": True,
+            "textDocumentSync": self.sync_type
         }
+        if self.use_signature_help:
+            server_capabilities["signatureHelpProvider"] = {
+                "triggerCharacters": ["(", ","]
+            }
+        return {"capabilities": server_capabilities}
         #     "workspaceSymbolProvider": True,
         #     "streaming": False,
         # }
 
+    def serve_workspace_symbol(self, request):
+        def map_types(type):
+            if type == 1:
+                return 2
+            elif type == 2:
+                return 6
+            elif type == 3:
+                return 12
+            elif type == 4:
+                return 5
+            elif type == 5:
+                return 11
+            elif type == 6:
+                return 13
+            elif type == 7:
+                return 6
+            else:
+                return 1
+
+        def add_children(mod_obj, query):
+            tmp_list = []
+            if mod_obj.get_type() == 1:
+                for child_obj in mod_obj.get_children():
+                    if not child_obj.name.lower().find(query) > -1:
+                        continue
+                    tmp_list.append(child_obj)
+            return tmp_list
+        matching_symbols = []
+        query = request["params"]["query"].lower()
+        for top_key, obj_packed in self.obj_tree.items():
+            top_obj = obj_packed[0]
+            top_uri = obj_packed[1]
+            if top_uri is None:
+                continue
+            candidates = add_children(top_obj, query)
+            if top_obj.name.lower().find(query) > -1:
+                candidates += [top_obj]
+            for candidate in candidates:
+                tmp_out = {}
+                tmp_out["name"] = candidate.name
+                tmp_out["kind"] = map_types(candidate.get_type())
+                sline = candidate.sline-1
+                eline = candidate.eline-1
+                # Set containing scope
+                if candidate.FQSN.find('::') > 0:
+                    tmp_list = candidate.FQSN.split("::")
+                    tmp_out["containerName"] = tmp_list[0]
+                # Set location
+                tmp_out["location"] = {
+                    "uri": path_to_uri(top_uri),
+                    "range": {
+                        "start": {"line": sline, "character": 0},
+                        "end": {"line": eline, "character": 0}
+                    }
+                }
+                matching_symbols.append(tmp_out)
+        return sorted(matching_symbols, key=lambda k: k['name'])
+
     def serve_document_symbols(self, request):
         def map_types(type):
             if type == 1:
@@ -509,6 +573,8 @@ class LangServer:
         # Add scopes to outline view
         test_output = []
         for scope in file_obj.get_scopes():
+            if (scope.name[0] == "#") or (scope.get_type() == 8):
+                continue
             scope_tree = scope.FQSN.split("::")
             if len(scope_tree) > 2:
                 if scope_tree[1].startswith("#gen_int"):
@@ -521,9 +587,7 @@ class LangServer:
             tmp_out["name"] = scope.name
             tmp_out["kind"] = scope_type
             sline = scope.sline-1
-            eline = sline
-            if scope.eline is not None:
-                eline = scope.eline-1
+            eline = scope.eline-1
             tmp_out["location"] = {
                 "uri": uri,
                 "range": {
@@ -565,6 +629,9 @@ class LangServer:
             else:
                 return type
 
+        def set_type_mask(def_value):
+            return [def_value if i < 8 else True for i in range(13)]
+
         def get_candidates(scope_list, var_prefix, inc_globals=True, public_only=False, abstract_only=False):
             #
             def child_candidates(scope, only_list=[], filter_public=True, req_abstract=False):
@@ -579,7 +646,7 @@ class LangServer:
                         if child.is_external_int():
                             tmp_list += child_candidates(child, only_list, filter_public)
                         else:
-                            if (nonly > 0) and (child.name not in only_list):
+                            if (nonly > 0) and (child.name.lower() not in only_list):
                                 continue
                             tmp_list.append(child)
                 return tmp_list
@@ -597,7 +664,7 @@ class LangServer:
             if inc_globals:
                 for key, obj in self.obj_tree.items():
                     var_list.append(obj[0])
-                var_list += self.intrinsics
+                var_list += self.intrinsic_funs
             # Filter by prefix if necessary
             if var_prefix == '':
                 return var_list
@@ -608,17 +675,27 @@ class LangServer:
                         tmp_list.append(var)
                 return tmp_list
 
-        def build_comp(candidate, name_only=False, name_replace=None):
+        def build_comp(candidate, name_only=False, name_replace=None, is_interface=False):
             comp_obj = {}
+            call_sig = None
             if name_only:
                 comp_obj["label"] = candidate.name
             else:
-                comp_obj["label"], snippet = candidate.get_snippet(name_replace)
+                comp_obj["label"] = candidate.name
+                if name_replace is not None:
+                    comp_obj["label"] = name_replace
+                call_sig, snippet = candidate.get_snippet(name_replace)
                 if snippet is not None:
+                    if self.use_signature_help and (not is_interface):
+                        arg_open = snippet.find('(')
+                        if arg_open > 0:
+                            snippet = snippet[:arg_open]
                     comp_obj["insertText"] = snippet
                     comp_obj["insertTextFormat"] = 2
             comp_obj["kind"] = map_types(candidate.get_type())
             comp_obj["detail"] = candidate.get_desc()
+            if call_sig is not None:
+                comp_obj["detail"] += ' ' + call_sig
             doc_str, _ = candidate.get_documentation()
             if doc_str is not None:
                 comp_obj["documentation"] = doc_str
@@ -670,6 +747,9 @@ class LangServer:
                         return 4, var_prefix, None
                     if PROCEDURE_STMNT_REGEX.match(test_str) is not None:
                         return 6, var_prefix, None
+            # Only thing on line?
+            if INT_STMNT_REGEX.match(line) is not None:
+                return 9, var_prefix, None
             # Default context
             if type_def:
                 return -1, var_prefix, None
@@ -732,17 +812,15 @@ class LangServer:
         # Setup based on context
         req_callable = False
         abstract_only = False
-        type_mask = [False for i in range(8)]
+        type_mask = set_type_mask(False)
         type_mask[1] = True
         type_mask[4] = True
         if line_context == 1:
             # Use statement module part (modules only)
             for key in self.obj_tree:
                 candidate = self.obj_tree[key][0]
-                candidate_type = candidate.get_type()
-                if candidate_type != 1:
-                    continue
-                if candidate.name.lower().startswith(var_prefix):
+                if (candidate.get_type() == 1) and \
+                   candidate.name.lower().startswith(var_prefix):
                     item_list.append(build_comp(candidate, name_only=True))
             req_dict["items"] = item_list
             return req_dict
@@ -763,12 +841,12 @@ class LangServer:
         elif line_context == 4:
             # Variable definition statement for user-defined type
             # (user-defined types only)
-            type_mask = [True for i in range(8)]
+            type_mask = set_type_mask(True)
             type_mask[4] = False
         elif line_context == 5:
             # Include statement (variables and user-defined types only)
             name_only = True
-            type_mask = [True for i in range(8)]
+            type_mask = set_type_mask(True)
             type_mask[4] = False
             type_mask[6] = False
         elif line_context == 6:
@@ -777,7 +855,7 @@ class LangServer:
             abstract_only = True
             include_globals = False
             name_only = True
-            type_mask = [True for i in range(8)]
+            type_mask = set_type_mask(True)
             type_mask[2] = False
             type_mask[3] = False
         elif line_context == 7:
@@ -795,11 +873,16 @@ class LangServer:
                 key_context = 2
             elif enc_scope_type == 4:
                 key_context = 3
-            for candidate in get_keywords(key_context):
+            for candidate in get_keywords(self.statements, self.keywords, key_context):
                 if candidate.name.lower().startswith(var_prefix):
                     item_list.append(build_comp(candidate))
             req_dict["items"] = item_list
             return req_dict
+        elif line_context == 9:
+            # First word -> default context plus Fortran statements
+            for candidate in get_keywords(self.statements, self.keywords, 0):
+                if candidate.name.lower().startswith(var_prefix):
+                    item_list.append(build_comp(candidate))
         for candidate in get_candidates(scope_list, var_prefix, include_globals, public_only, abstract_only):
             # Skip module names (only valid in USE)
             candidate_type = candidate.get_type()
@@ -815,7 +898,7 @@ class LangServer:
                     if tmp_list.count(tmp_text) > 0:
                         continue
                     tmp_list.append(tmp_text)
-                    item_list.append(build_comp(member, name_replace=candidate.name))
+                    item_list.append(build_comp(member, name_replace=candidate.name, is_interface=True))
                 continue
             #
             item_list.append(build_comp(candidate, name_only=name_only))
@@ -866,13 +949,136 @@ class LangServer:
             key = def_name.lower()
             if key in self.obj_tree:
                 return self.obj_tree[key][0]
-            for obj in self.intrinsics:
+            for obj in self.intrinsic_funs:
                 if obj.name.lower() == key:
                     return obj
         else:
             return var_obj
         return None
 
+    def serve_signature(self, request):
+        def get_sub_name(line):
+            nLine = len(line)
+            line_grouped = tokenize_line(line)
+            if len(line_grouped) < 2:
+                return None, None, None
+            lowest_level = -1
+            for i, level in enumerate(line_grouped):
+                if level[-1][0][-1][-1] == nLine:
+                    lowest_level = i
+            if lowest_level > 0:
+                arg_string = ''
+                for char_group in line_grouped[lowest_level]:
+                    arg_string += char_group[-1]
+                return line_grouped[lowest_level-1][0][1].strip(), arg_string.split(','), \
+                    line_grouped[lowest_level-1][0][0][-1][1]
+            else:
+                return None, None, None
+
+        def check_optional(arg, params):
+            opt_split = arg.split("=")
+            if len(opt_split) > 1:
+                opt_arg = opt_split[0].strip().lower()
+                for i, param in enumerate(params):
+                    param_split = param["label"].split("=")[0]
+                    if param_split.lower() == opt_arg:
+                        return i
+            return None
+        # Get parameters from request
+        req_dict = {"signatures": []}
+        params = request["params"]
+        uri = params["textDocument"]["uri"]
+        path = path_from_uri(uri)
+        if path not in self.workspace:
+            return req_dict
+        # Check line
+        sig_line = params["position"]["line"]
+        sig_char = params["position"]["character"]
+        # Get full line (and possible continuations) from file
+        curr_line, sig_char = get_line(sig_line, sig_char, self.workspace[path])
+        if curr_line is None:
+            return req_dict
+        # Test if scope declaration or end statement
+        if SCOPE_DEF_REGEX.match(curr_line) or END_REGEX.match(curr_line):
+            return req_dict
+        is_member = False
+        try:
+            line_prefix = curr_line[:sig_char].lower()
+            # Ignore for comment lines
+            comm_start = detect_comment_start(line_prefix, self.workspace[path]["fixed"])
+            if (comm_start >= 0) or (line_prefix[0] == '#'):
+                return req_dict
+            # Ignore string literals
+            if (line_prefix.count("'") % 2 == 1) or \
+               (line_prefix.count('"') % 2 == 1):
+                return req_dict
+            sub_name, arg_strings, sub_end = get_sub_name(line_prefix)
+            var_stack = get_var_stack(sub_name)
+            is_member = (len(var_stack) > 1)
+        except:
+            return req_dict
+        #
+        file_obj = self.workspace[path]["ast"]
+        curr_scope = file_obj.get_inner_scope(sig_line+1)
+        # Traverse type tree if necessary
+        if is_member:
+            type_scope = climb_type_tree(var_stack, curr_scope, self.obj_tree)
+            # Set enclosing type as scope
+            if type_scope is None:
+                curr_scope = None
+            else:
+                curr_scope = type_scope
+        sub_name = var_stack[-1]
+        # Find in available scopes
+        var_obj = None
+        if curr_scope is not None:
+            var_obj, enc_scope = find_in_scope(curr_scope, sub_name, self.obj_tree)
+        # Search in global scope
+        if var_obj is None:
+            key = sub_name.lower()
+            if key in self.obj_tree:
+                var_obj = self.obj_tree[key][0]
+            else:
+                for obj in self.intrinsic_funs:
+                    if obj.name.lower() == key:
+                        var_obj = obj
+                        break
+        # Check keywords
+        if (var_obj is None) and (INT_STMNT_REGEX.match(line_prefix[:sub_end]) is not None):
+            key = sub_name.lower()
+            for candidate in get_keywords(self.statements, self.keywords, 0):
+                if candidate.name.lower() == key:
+                    var_obj = candidate
+                    break
+        if var_obj is None:
+            return req_dict
+        # Build signature
+        label, doc_str, params = var_obj.get_signature()
+        if label is None:
+            return req_dict
+        # Find current parameter by index or by
+        # looking at last arg with optional name
+        param_num = len(arg_strings)-1
+        opt_num = check_optional(arg_strings[-1], params)
+        if opt_num is None:
+            if len(arg_strings) > 1:
+                opt_num = check_optional(arg_strings[-2], params)
+                if opt_num is not None:
+                    param_num = opt_num + 1
+        else:
+            param_num = opt_num
+        signature = {
+            "label": label,
+            "parameters": params
+        }
+        if doc_str is not None:
+            signature["documentation"] = doc_str
+        req_dict = {
+            "signatures": [signature],
+            "activeParameter": param_num
+        }
+        return req_dict
+
     def serve_references(self, request):
         # Get parameters from request
         params = request["params"]
@@ -945,6 +1151,14 @@ class LangServer:
             return None
 
     def serve_hover(self, request):
+        def create_hover(string, highlight):
+            if highlight:
+                return {
+                    "language": "fortran90",
+                    "value": string
+                }
+            else:
+                return string
         # Get parameters from request
         params = request["params"]
         uri = params["textDocument"]["uri"]
@@ -959,24 +1173,40 @@ class LangServer:
         # Construct hover information
         if var_obj is not None:
             var_type = var_obj.get_type()
-            # Currently only show for subroutines
+            hover_str = None
             if var_type == 2 or var_type == 3:
                 hover_str, highlight = var_obj.get_documentation(long=True)
-                if hover_str is not None:
-                    if highlight:
-                        return {
-                            "contents": {
-                                "language": "fortran",
-                                "value": hover_str
-                            }
-                        }
-                    else:
-                        return {
-                            "contents": hover_str
-                        }
+            elif var_type == 5:
+                hover_array = []
+                for member in var_obj.mems:
+                    hover_str, highlight = member.get_documentation(long=True)
+                    if hover_str is not None:
+                        hover_array.append(create_hover(hover_str, highlight))
+                return {"contents": hover_array}
+            elif var_type == 6 and self.variable_hover:
+                hover_str, highlight = var_obj.get_documentation()
+            #
+            if hover_str is not None:
+                return {"contents": create_hover(hover_str, highlight)}
         else:
             return None
 
+    def send_diagnostics(self, uri):
+        diag_results, diag_exp = self.get_diagnostics(uri)
+        if diag_results is not None:
+            self.conn.send_notification("textDocument/publishDiagnostics", {
+                "uri": uri,
+                "diagnostics": diag_results
+            })
+        elif diag_exp is not None:
+            self.conn.write_error(
+                -1,
+                code=-32603,
+                message=str(diag_exp),
+                data={
+                    "traceback": traceback.format_exc(),
+                })
+
     def get_diagnostics(self, uri):
         filepath = path_from_uri(uri)
         if filepath in self.workspace:
@@ -985,18 +1215,10 @@ class LangServer:
             try:
                 diags = file_obj.check_file(self.obj_tree, file_contents)
             except Exception as e:
-                self.conn.write_error(
-                    -1,
-                    code=-32603,
-                    message=str(e),
-                    data={
-                        "traceback": traceback.format_exc(),
-                    })
+                return None, e
             else:
-                self.conn.send_notification("textDocument/publishDiagnostics", {
-                    "uri": uri,
-                    "diagnostics": diags
-                })
+                return diags, None
+        return None, None
 
     def serve_onChange(self, request):
         # Update workspace from file sent by editor
@@ -1005,7 +1227,10 @@ class LangServer:
         path = path_from_uri(uri)
         # Update file contents with changes
         if self.sync_type == 1:
-            new_contents = params["contentChanges"][0]["text"].splitlines()
+            file_text = params["contentChanges"][0]["text"]
+            if not PY3K:
+                file_text = file_text.encode('utf-8')
+            new_contents = file_text.splitlines()
         else:
             if path in self.workspace:
                 new_contents = self.workspace[path]["contents"]
@@ -1015,9 +1240,11 @@ class LangServer:
                         new_contents, line_tmp = apply_change(old_contents, change)
                 except:
                     self.post_message('Change request failed for file "{0}": Could not apply change'.format(path))
+                    log.error('Change request failed for file "%s": Could not apply change', path, exc_info=True)
                     return
             else:
                 self.post_message('Change request failed for unknown file "{0}"'.format(path))
+                log.error('Change request failed for unknown file "%s"', path)
                 return
         # Parse newly updated file
         err_str = self.update_workspace_file(new_contents, path, update_links=True)
@@ -1033,23 +1260,29 @@ class LangServer:
         #     self.obj_tree[key][0].resolve_inherit(self.obj_tree)
         #     self.obj_tree[key][0].resolve_link(self.obj_tree)
 
-    def serve_onSave(self, request):
+    def serve_onClose(self, request):
+        self.serve_onSave(request, test_exist=True)
+
+    def serve_onSave(self, request, test_exist=False):
         # Update workspace from file on disk
         params = request["params"]
         uri = params["textDocument"]["uri"]
         filepath = path_from_uri(uri)
+        if test_exist and (not os.path.isfile(filepath)):
+            return
         err_str = self.add_file(filepath)
         if err_str is not None:
             self.post_message('Save request failed for file "{0}": {1}'.format(filepath, err_str))
             return
-        # Update include statements
-        for path, file_obj in self.workspace.items():
-            file_obj["ast"].resolve_includes(self.workspace)
+        # Update include statements linking to this file
+        for tmp_path, file_obj in self.workspace.items():
+            file_obj["ast"].resolve_includes(self.workspace, path=filepath)
+        self.workspace[filepath]["ast"].resolve_includes(self.workspace)
         # Update inheritance
         for key in self.obj_tree:
             self.obj_tree[key][0].resolve_inherit(self.obj_tree)
             self.obj_tree[key][0].resolve_link(self.obj_tree)
-        self.get_diagnostics(uri)
+        self.send_diagnostics(uri)
 
     def add_file(self, filepath):
         # Read and add file from disk
@@ -1064,6 +1297,7 @@ class LangServer:
             fixed_flag = detect_fixed_format(contents_split)
             ast_new = process_file(contents_split, True, filepath, fixed_flag)
         except:
+            log.error("Error while parsing file %s", filepath, exc_info=True)
             return 'Error during parsing'  # Error during parsing
         # Remove old objects from tree
         if filepath in self.workspace:
@@ -1097,7 +1331,13 @@ class LangServer:
                     filepath = os.path.join(mod_dir, filename)
                     if self.excl_paths.count(filepath) > 0:
                         continue
-                    file_list.append(filepath)
+                    inc_file = True
+                    for excl_suffix in self.excl_suffixes:
+                        if filepath.endswith(excl_suffix):
+                            inc_file = False
+                            break
+                    if inc_file:
+                        file_list.append(filepath)
         # Process files
         from multiprocessing import Pool
         pool = Pool(processes=4)
