@@ -1,12 +1,15 @@
 import copy
 import re
 import os
+from fortls.jsonrpc import path_to_uri
 WORD_REGEX = re.compile(r'[a-z_][a-z0-9_]*', re.I)
 CLASS_VAR_REGEX = re.compile(r'(TYPE|CLASS)[ ]*\(', re.I)
 
 
 def map_keywords(keywords):
     modifiers = []
+    pass_name = None
+    dim_str = None
     for key in keywords:
         key_lower = key.lower()
         if key_lower == 'pointer':
@@ -21,20 +24,38 @@ def map_keywords(keywords):
             modifiers.append(5)
         elif key_lower == 'nopass':
             modifiers.append(6)
-        elif key_lower == 'intent(in)':
+        elif key_lower == 'target':
             modifiers.append(7)
-        elif key_lower == 'intent(out)':
+        elif key_lower == 'save':
             modifiers.append(8)
-        elif key_lower == 'intent(inout)':
+        elif key_lower == 'parameter':
             modifiers.append(9)
+        elif key_lower == 'contiguous':
+            modifiers.append(10)
+        elif key_lower == 'deferred':
+            modifiers.append(11)
+        elif key_lower == 'intent(in)':
+            modifiers.append(101)
+        elif key_lower == 'intent(out)':
+            modifiers.append(102)
+        elif key_lower == 'intent(inout)':
+            modifiers.append(103)
         elif key_lower.startswith('dimension'):
-            ndims = key_lower.count(':')
-            modifiers.append(20+ndims)
+            i1 = key_lower.find('(')
+            i2 = key_lower.find(')')
+            if i1 > -1 and i2 > i1:
+                dim_str = key_lower[i1+1:i2]
+                modifiers.append(20)
+        elif key_lower.startswith('pass'):
+            i1 = key_lower.find('(')
+            i2 = key_lower.find(')')
+            if i1 > -1 and i2 > i1:
+                pass_name = key_lower[i1+1:i2]
     modifiers.sort()
-    return modifiers
+    return modifiers, dim_str, pass_name
 
 
-def get_keywords(modifiers):
+def get_keywords(modifiers, dim_str=None):
     mod_strings = []
     for modifier in modifiers:
         if modifier == 1:
@@ -50,15 +71,22 @@ def get_keywords(modifiers):
         elif modifier == 6:
             mod_strings.append('NOPASS')
         elif modifier == 7:
-            mod_strings.append('INTENT(IN)')
+            mod_strings.append('TARGET')
         elif modifier == 8:
-            mod_strings.append('INTENT(OUT)')
+            mod_strings.append('SAVE')
         elif modifier == 9:
+            mod_strings.append('PARAMETER')
+        elif modifier == 10:
+            mod_strings.append('CONTIGUOUS')
+        elif modifier == 11:
+            mod_strings.append('DEFERRED')
+        elif modifier == 101:
+            mod_strings.append('INTENT(IN)')
+        elif modifier == 102:
+            mod_strings.append('INTENT(OUT)')
+        elif modifier == 103:
             mod_strings.append('INTENT(INOUT)')
-        elif modifier > 20:
-            dim_str = ":"
-            for i in range(modifier-21):
-                dim_str += ",:"
+        elif (modifier == 20) and (dim_str is not None):
             mod_strings.append('DIMENSION({0})'.format(dim_str))
     return mod_strings
 
@@ -167,7 +195,7 @@ class fortran_scope:
     def base_setup(self, file_obj, sline, name, enc_scope=None):
         self.file = file_obj
         self.sline = sline
-        self.eline = None
+        self.eline = sline
         self.name = name
         self.children = []
         self.members = []
@@ -228,25 +256,26 @@ class fortran_scope:
     def get_desc(self):
         return 'unknown'
 
-    def get_snippet(self, name_replace=None, drop_arg=None):
-        if name_replace is not None:
-            return name_replace, None
-        return self.name, None
+    def get_snippet(self, name_replace=None, drop_arg=-1):
+        return None, None
 
     def get_documentation(self, long=False):
         return None, False
 
+    def get_signature(self):
+        return None, None, None
+
     def get_children(self, public_only=False):
         if public_only:
             pub_children = []
             for child in self.children:
-                if child.name.startswith("#GEN_INT"):
-                    pub_children.append(child)
-                    continue
                 if child.vis < 0:
                     continue
                 if (self.def_vis < 0) and (child.vis <= 0):
                     continue
+                if child.name.startswith("#GEN_INT"):
+                    pub_children.append(child)
+                    continue
                 pub_children.append(child)
             return pub_children
         else:
@@ -270,6 +299,9 @@ class fortran_scope:
     def is_abstract(self):
         return False
 
+    def req_named_end(self):
+        return False
+
     def end(self, line_number):
         self.eline = line_number
 
@@ -288,17 +320,18 @@ class fortran_scope:
                 errors.append([0, line_number, i0, i1, child.name])
             else:
                 FQSN_list.append(child.FQSN)
-            # Check for masking from parent scope in subroutines and functions
-            if self.parent is not None and (self.get_type() == 2 or self.get_type() == 3):
+            # Check for masking from parent scope in subroutines, functions, and blocks
+            if (self.parent is not None) and (self.get_type() in (2, 3, 8)):
                 parent_var, parent_scope = \
                     find_in_scope(self.parent, child.name, obj_tree)
                 if parent_var is not None:
                     # Ignore if function return variable
-                    if self.get_type() == 3 and parent_var.FQSN == self.FQSN:
+                    if (self.get_type() == 3) and (parent_var.FQSN == self.FQSN):
                         continue
                     line_number = child.sline - 1
                     i0, i1 = find_word_in_line(file_contents[line_number].lower(), child.name.lower())
-                    errors.append([1, line_number, i0, i1, child.name])
+                    errors.append([1, line_number, i0, i1, child.name,
+                                   parent_var.file.path, parent_var.sline - 1])
         return errors
 
     def check_use(self, obj_tree, file_contents):
@@ -314,6 +347,9 @@ class fortran_scope:
                 errors.append([line_number, i0, i0+len(use_mod), use_mod])
         return errors
 
+    def check_valid_parent(self):
+        return True
+
 
 class fortran_module(fortran_scope):
     def get_type(self):
@@ -322,6 +358,11 @@ class fortran_module(fortran_scope):
     def get_desc(self):
         return 'MODULE'
 
+    def check_valid_parent(self):
+        if self.parent is not None:
+            return False
+        return True
+
 
 class fortran_program(fortran_module):
     def get_desc(self):
@@ -375,7 +416,8 @@ class fortran_submodule(fortran_module):
 class fortran_subroutine(fortran_scope):
     def __init__(self, file_obj, line_number, name, enc_scope=None, args="", mod_sub=False):
         self.base_setup(file_obj, line_number, name, enc_scope)
-        self.args = args
+        self.args = args.replace(' ', '').lower()
+        self.args_snip = self.args
         self.arg_objs = []
         self.in_children = []
         self.mod_scope = mod_sub
@@ -389,6 +431,7 @@ class fortran_subroutine(fortran_scope):
     def copy_interface(self, copy_source):
         # Copy arguments
         self.args = copy_source.args
+        self.args_snip = copy_source.args_snip
         self.arg_objs = copy_source.arg_objs
         # Get current fields
         child_names = []
@@ -397,6 +440,8 @@ class fortran_subroutine(fortran_scope):
         # Import arg_objs from copy object
         self.in_children = []
         for child in copy_source.arg_objs:
+            if child is None:
+                continue
             if child.name.lower() not in child_names:
                 self.in_children.append(child)
 
@@ -406,20 +451,24 @@ class fortran_subroutine(fortran_scope):
         return tmp_list
 
     def resolve_arg_link(self, obj_tree):
-        self.arg_objs = []
-        arg_list = self.args.replace(' ', '').lower().split(',')
+        arg_list = self.args.replace(' ', '').split(',')
+        self.arg_objs = [None for arg in arg_list]
+        check_objs = self.children
         for child in self.children:
+            if child.is_external_int():
+                check_objs += child.get_children()
+        for child in check_objs:
             ind = -1
             for i, arg in enumerate(arg_list):
                 if arg == child.name.lower():
                     ind = i
                     break
             if ind >= 0:
-                self.arg_objs.append(child)
+                self.arg_objs[ind] = child
                 if child.is_optional():
                     arg_list[ind] = "{0}={0}".format(arg_list[ind])
             child.resolve_link(obj_tree)
-        self.args = ",".join(arg_list)
+        self.args_snip = ",".join(arg_list)
 
     def resolve_link(self, obj_tree):
         self.resolve_arg_link(obj_tree)
@@ -427,13 +476,10 @@ class fortran_subroutine(fortran_scope):
     def get_type(self):
         return 2
 
-    def get_snippet(self, name_replace=None, drop_arg=None):
-        arg_list = self.args.split(",")
-        if drop_arg is not None:
-            if len(arg_list) > 1:
-                arg_list = arg_list[1:]
-            else:
-                arg_list = []
+    def get_snippet(self, name_replace=None, drop_arg=-1):
+        arg_list = self.args_snip.split(",")
+        if (drop_arg >= 0) and (drop_arg < len(arg_list)):
+            del arg_list[drop_arg]
         arg_snip = None
         if len(arg_list) > 0:
             place_holders = []
@@ -443,8 +489,8 @@ class fortran_subroutine(fortran_scope):
                     place_holders.append("{1}=${{{0}:{2}}}".format(i+1, opt_split[0], opt_split[1]))
                 else:
                     place_holders.append("${{{0}:{1}}}".format(i+1, arg))
-            arg_str = "({0})".format(",".join(arg_list))
-            arg_snip = "({0})".format(",".join(place_holders))
+            arg_str = "({0})".format(", ".join(arg_list))
+            arg_snip = "({0})".format(", ".join(place_holders))
         else:
             arg_str = "()"
         name = self.name
@@ -459,28 +505,50 @@ class fortran_subroutine(fortran_scope):
         return 'SUBROUTINE'
 
     def get_documentation(self, long=False):
+        sub_sig, _ = self.get_snippet()
         if long:
-            skip_arg = False
-            hover_str, _ = self.get_snippet()
-            hover_str += "\n"
+            hover_array = ["SUBROUTINE " + sub_sig]
             for arg_obj in self.arg_objs:
-                if skip_arg:
-                    skip_arg = False
+                if arg_obj is None:
                     continue
                 arg_doc, _ = arg_obj.get_documentation()
-                tmp_str = "  " + arg_doc
-                tmp_str += " :: {0}".format(arg_obj.name)
-                hover_str = hover_str + tmp_str + "\n"
-            return hover_str, True
+                hover_array.append(" {0} :: {1}".format(arg_doc, arg_obj.name))
+            return "\n".join(hover_array), True
         else:
-            return None, False
+            return "SUBROUTINE " + sub_sig, False
+
+    def get_signature(self):
+        arg_sigs = []
+        arg_list = self.args.split(",")
+        for i, arg_obj in enumerate(self.arg_objs):
+            if arg_obj is None:
+                arg_sigs.append({"label": arg_list[i]})
+            else:
+                if arg_obj.is_optional():
+                    label = "{0}={0}".format(arg_obj.name.lower())
+                else:
+                    label = arg_obj.name.lower()
+                arg_sigs.append({
+                    "label": label,
+                    "documentation": arg_obj.get_documentation()[0]
+                })
+        call_sig, _ = self.get_snippet()
+        return call_sig, None, arg_sigs
+
+    def check_valid_parent(self):
+        if self.parent is not None:
+            parent_type = self.parent.get_type()
+            if (parent_type == 4) or (parent_type >= 8):
+                return False
+        return True
 
 
 class fortran_function(fortran_subroutine):
     def __init__(self, file_obj, line_number, name, enc_scope=None, args="",
                  mod_fun=False, return_type=None, result_var=None):
         self.base_setup(file_obj, line_number, name, enc_scope)
-        self.args = args
+        self.args = args.replace(' ', '').lower()
+        self.args_snip = self.args
         self.arg_objs = []
         self.in_children = []
         self.mod_scope = mod_fun
@@ -488,14 +556,16 @@ class fortran_function(fortran_subroutine):
         self.result_obj = None
         if return_type is not None:
             self.return_type = return_type[0]
-            self.modifiers = map_keywords(return_type[1])
+            self.modifiers, self.dim_str, _ = map_keywords(return_type[1])
         else:
             self.return_type = None
             self.modifiers = []
+            self.dim_str = None
 
     def copy_interface(self, copy_source):
         # Copy arguments and returns
         self.args = copy_source.args
+        self.args_snip = copy_source.args_snip
         self.arg_objs = copy_source.arg_objs
         self.result_var = copy_source.result_var
         self.result_obj = copy_source.result_obj
@@ -506,6 +576,8 @@ class fortran_function(fortran_subroutine):
         # Import arg_objs from copy object
         self.in_children = []
         for child in copy_source.arg_objs:
+            if child is None:
+                continue
             if child.name.lower() not in child_names:
                 self.in_children.append(child)
         if copy_source.result_obj is not None:
@@ -526,14 +598,32 @@ class fortran_function(fortran_subroutine):
     def get_desc(self):
         # desc = None
         if self.result_obj is not None:
-            return self.result_obj.get_desc()
+            return self.result_obj.get_desc() + ' FUNCTION'
         if self.return_type is not None:
-            return self.return_type
+            return self.return_type + ' FUNCTION'
         return 'FUNCTION'
 
     def is_callable(self):
         return False
 
+    def get_documentation(self, long=False):
+        fun_sig, _ = self.get_snippet()
+        fun_return = ""
+        if self.result_obj is not None:
+            fun_return, _ = self.result_obj.get_documentation()
+        if self.return_type is not None:
+            fun_return = self.return_type
+        if long:
+            hover_array = ["{0} FUNCTION {1}".format(fun_return, fun_sig)]
+            for arg_obj in self.arg_objs:
+                if arg_obj is None:
+                    continue
+                arg_doc, _ = arg_obj.get_documentation()
+                hover_array.append(" {0} :: {1}".format(arg_doc, arg_obj.name))
+            return "\n".join(hover_array), True
+        else:
+            return "{0} FUNCTION {1}".format(fun_return, fun_sig) + fun_sig, False
+
 
 class fortran_type(fortran_scope):
     def __init__(self, file_obj, line_number, name, modifiers, enc_scope=None):
@@ -578,6 +668,100 @@ class fortran_type(fortran_scope):
                 if child.name.lower() not in child_names:
                     self.in_children.append(child)
 
+    def check_valid_parent(self):
+        if self.parent is None:
+            return False
+        else:
+            parent_type = self.parent.get_type()
+            if (parent_type == 4) or (parent_type >= 8):
+                return False
+        return True
+
+
+class fortran_block(fortran_scope):
+    def __init__(self, file_obj, line_number, name, enc_scope=None):
+        self.base_setup(file_obj, line_number, name, enc_scope)
+
+    def get_type(self):
+        return 8
+
+    def get_desc(self):
+        return 'BLOCK'
+
+    def get_children(self, public_only=False):
+        return self.children
+
+    def req_named_end(self):
+        return True
+
+
+class fortran_do(fortran_block):
+    def __init__(self, file_obj, line_number, name, enc_scope=None):
+        self.base_setup(file_obj, line_number, name, enc_scope)
+
+    def get_type(self):
+        return 10
+
+    def get_desc(self):
+        return 'DO'
+
+
+class fortran_where(fortran_block):
+    def __init__(self, file_obj, line_number, name, enc_scope=None):
+        self.base_setup(file_obj, line_number, name, enc_scope)
+
+    def get_type(self):
+        return 11
+
+    def get_desc(self):
+        return 'WHERE'
+
+
+class fortran_if(fortran_block):
+    def __init__(self, file_obj, line_number, name, enc_scope=None):
+        self.base_setup(file_obj, line_number, name, enc_scope)
+
+    def get_type(self):
+        return 12
+
+    def get_desc(self):
+        return 'IF'
+
+
+class fortran_associate(fortran_block):
+    def __init__(self, file_obj, line_number, name, enc_scope=None):
+        self.base_setup(file_obj, line_number, name, enc_scope)
+
+    def get_type(self):
+        return 13
+
+    def get_desc(self):
+        return 'ASSOCIATE'
+
+
+class fortran_select(fortran_block):
+    def __init__(self, file_obj, line_number, name, select_info, enc_scope=None):
+        self.base_setup(file_obj, line_number, name, enc_scope)
+        self.type = select_info[0]
+        self.binding_name = None
+        self.bound_var = None
+        self.binding_type = None
+        if self.type == 2:
+            binding_split = select_info[1].split('=>')
+            if len(binding_split) == 1:
+                self.bound_var = binding_split[0].strip()
+            elif len(binding_split) == 2:
+                self.binding_name = binding_split[0].strip()
+                self.bound_var = binding_split[1].strip()
+        elif self.type == 3:
+            self.binding_type = select_info[1]
+
+    def get_type(self):
+        return 9
+
+    def get_desc(self):
+        return 'SELECT'
+
 
 class fortran_int(fortran_scope):
     def __init__(self, file_obj, line_number, name, enc_scope=None, abstract=False):
@@ -615,14 +799,22 @@ class fortran_int(fortran_scope):
 
 class fortran_obj:
     def __init__(self, file_obj, line_number, name, var_desc, modifiers,
-                 enc_scope=None, link_obj=None):
+                 dim_str, enc_scope=None, link_obj=None):
+        self.base_setup(file_obj, line_number, name, var_desc, modifiers,
+                        dim_str, enc_scope, link_obj)
+
+    def base_setup(self, file_obj, line_number, name, var_desc, modifiers,
+                   dim_str, enc_scope, link_obj):
         self.file = file_obj
         self.sline = line_number
+        self.eline = line_number
         self.name = name
         self.desc = var_desc
         self.modifiers = modifiers
+        self.dim_str = dim_str
         self.callable = (CLASS_VAR_REGEX.match(var_desc) is not None)
         self.children = []
+        self.use = []
         self.vis = 0
         self.parent = None
         self.link_obj = None
@@ -655,13 +847,9 @@ class fortran_obj:
         if self.link_name is None:
             return
         if self.parent is not None:
-            parent = self.parent
-            while(parent is not None):
-                for child in parent.children:
-                    if child.name.lower() == self.link_name:
-                        self.link_obj = child
-                        return
-                parent = parent.parent
+            link_obj, _ = find_in_scope(self.parent, self.link_name, obj_tree)
+            if link_obj is not None:
+                self.link_obj = link_obj
 
     def set_visibility(self, new_vis):
         self.vis = new_vis
@@ -678,32 +866,31 @@ class fortran_obj:
         # Normal variable
         return self.desc
 
-    def set_dim(self, ndim):
-        for (i, modifier) in enumerate(self.modifiers):
-            if modifier > 20:
-                self.modifiers[i] = ndim+20
-                return
-        self.modifiers.append(ndim+20)
+    def set_dim(self, dim_str):
+        self.dim_str = dim_str
+        if 20 not in self.modifiers:
+            self.modifiers.append(20)
+            self.modifiers.sort()
 
-    def get_snippet(self, name_replace=None, drop_arg=None):
+    def get_snippet(self, name_replace=None, drop_arg=-1):
         name = self.name
         if name_replace is not None:
             name = name_replace
         if self.link_obj is not None:
             return self.link_obj.get_snippet(name, drop_arg)
         # Normal variable
-        return name, None
+        return None, None
 
     def get_documentation(self, long=False):
-        if self.link_obj is not None:
-            return self.link_obj.get_documentation()
-        #
         doc_str = self.desc
         if len(self.modifiers) > 0:
             doc_str += ", "
-            doc_str += ", ".join(get_keywords(self.modifiers))
+            doc_str += ", ".join(get_keywords(self.modifiers, self.dim_str))
         return doc_str, True
 
+    def get_signature(self):
+        return None, None, None
+
     def get_children(self, public_only=False):
         return []
 
@@ -727,18 +914,30 @@ class fortran_obj:
 
 
 class fortran_meth(fortran_obj):
-    def get_snippet(self, name_replace=None, drop_arg=None):
-        if self.modifiers.count(6) > 0:
-            nopass = True
-        else:
-            nopass = False
-        #
+    def __init__(self, file_obj, line_number, name, var_desc, modifiers,
+                 enc_scope=None, link_obj=None, pass_name=None):
+        self.base_setup(file_obj, line_number, name, var_desc, modifiers,
+                        None, enc_scope, link_obj)
+        self.drop_arg = -1
+        self.pass_name = pass_name
+        if link_obj is None:
+            open_paren = var_desc.find('(')
+            close_paren = var_desc.find(')')
+            if (open_paren > 0) and (close_paren > open_paren):
+                self.link_name = var_desc[open_paren+1:close_paren].lower()
+
+    def set_parent(self, parent_obj):
+        self.parent = parent_obj
+        if (self.parent.get_type() == 4) and (self.modifiers.count(6) == 0):
+            self.drop_arg = 0
+
+    def get_snippet(self, name_replace=None, drop_arg=-1):
         name = self.name
         if name_replace is not None:
             name = name_replace
         if self.link_obj is not None:
-            return self.link_obj.get_snippet(name, nopass)
-        return name, None
+            return self.link_obj.get_snippet(name, self.drop_arg)
+        return None, None
 
     def get_type(self):
         if self.link_obj is not None:
@@ -748,31 +947,70 @@ class fortran_meth(fortran_obj):
 
     def get_documentation(self, long=False):
         if long:
-            skip_arg = False
-            hover_str, _ = self.get_snippet()
-            hover_str += "\n"
-            if self.modifiers.count(6) == 0:
-                skip_arg = True
-            var_obj = self.link_obj
-            for arg_obj in var_obj.arg_objs:
-                if skip_arg:
-                    skip_arg = False
-                    continue
-                arg_doc, _ = arg_obj.get_documentation()
-                tmp_str = "  " + arg_doc
-                tmp_str += " :: {0}".format(arg_obj.name)
-                hover_str = hover_str + tmp_str + "\n"
+            sub_sig, _ = self.get_snippet()
+            hover_str = "{0} {1}\n".format(self.get_desc(), sub_sig)
+            if self.link_obj is not None:
+                link_hover, _ = self.link_obj.get_documentation(long=True)
+                hover_split = link_hover.splitlines()
+                call_sig = hover_split[0]
+                paren_start = call_sig.rfind('(')
+                link_name_len = len(self.link_obj.name)
+                call_sig = call_sig[:paren_start-link_name_len] + self.name + call_sig[paren_start:]
+                paren_start += len(self.name) - link_name_len
+                hover_split = hover_split[1:]
+                if (self.drop_arg >= 0) and (self.drop_arg < len(hover_split)):
+                    paren_end = call_sig.rfind(')')
+                    args = call_sig[paren_start+1:paren_end].split(',')
+                    del args[self.drop_arg]
+                    del hover_split[self.drop_arg]
+                    call_sig = call_sig[:paren_start] + '(' + (','.join(args)).strip() + ')'
+                hover_str = '\n'.join([call_sig] + hover_split)
             return hover_str, True
         else:
-            if self.link_obj is not None:
-                return self.link_obj.get_documentation()
-            #
             doc_str = self.desc
             if len(self.modifiers) > 0:
                 doc_str += ", "
                 doc_str += ", ".join(get_keywords(self.modifiers))
             return doc_str, True
 
+    def get_signature(self):
+        arg_sigs = []
+        var_obj = self.link_obj
+        arg_list = var_obj.args.split(",")
+        for i, arg_obj in enumerate(var_obj.arg_objs):
+            if self.drop_arg == i:
+                continue
+            if arg_obj is None:
+                arg_sigs.append({"label": arg_list[i]})
+            else:
+                if arg_obj.is_optional():
+                    label = "{0}={0}".format(arg_obj.name.lower())
+                else:
+                    label = arg_obj.name.lower()
+                arg_sigs.append({
+                    "label": label,
+                    "documentation": arg_obj.get_documentation()[0]
+                })
+        call_sig, _ = self.get_snippet()
+        return call_sig, None, arg_sigs
+
+    def resolve_link(self, obj_tree):
+        if self.link_name is None:
+            return
+        if self.parent is not None:
+            if self.parent.get_type() == 4:
+                link_obj, _ = find_in_scope(self.parent.parent, self.link_name, obj_tree)
+            else:
+                link_obj, _ = find_in_scope(self.parent, self.link_name, obj_tree)
+            if link_obj is not None:
+                self.link_obj = link_obj
+                if self.pass_name is not None:
+                    self.pass_name = self.pass_name.lower()
+                    for i, arg in enumerate(link_obj.args_snip.split(',')):
+                        if arg.lower() == self.pass_name:
+                            self.drop_arg = i
+                            break
+
     def is_callable(self):
         return True
 
@@ -789,9 +1027,11 @@ class fortran_file:
         self.end_stack = []
         self.pp_if = []
         self.include_stmnts = []
+        self.end_errors = []
         self.none_scope = None
+        self.inc_scope = None
         self.current_scope = None
-        self.END_REGEX = None
+        self.END_SCOPE_WORD = None
         self.enc_scope_name = None
 
     def create_none_scope(self):
@@ -805,11 +1045,8 @@ class fortran_file:
             return None
         return self.current_scope.FQSN
 
-    def add_scope(self, new_scope, END_SCOPE_REGEX, hidden=False, exportable=True, req_container=False):
-        if hidden:
-            self.variable_list.append(new_scope)
-        else:
-            self.scope_list.append(new_scope)
+    def add_scope(self, new_scope, END_SCOPE_WORD, exportable=True, req_container=False):
+        self.scope_list.append(new_scope)
         if self.current_scope is None:
             if req_container:
                 self.create_none_scope()
@@ -822,10 +1059,10 @@ class fortran_file:
         else:
             self.current_scope.add_child(new_scope)
             self.scope_stack.append(self.current_scope)
-        if self.END_REGEX is not None:
-            self.end_stack.append(self.END_REGEX)
+        if self.END_SCOPE_WORD is not None:
+            self.end_stack.append(self.END_SCOPE_WORD)
         self.current_scope = new_scope
-        self.END_REGEX = END_SCOPE_REGEX
+        self.END_SCOPE_WORD = END_SCOPE_WORD
         self.enc_scope_name = self.get_enc_scope_name()
 
     def end_scope(self, line_number):
@@ -835,9 +1072,9 @@ class fortran_file:
         else:
             self.current_scope = None
         if len(self.end_stack) > 0:
-            self.END_REGEX = self.end_stack.pop()
+            self.END_SCOPE_WORD = self.end_stack.pop()
         else:
-            self.END_REGEX = None
+            self.END_SCOPE_WORD = None
         self.enc_scope_name = self.get_enc_scope_name()
 
     def add_variable(self, new_var):
@@ -884,11 +1121,11 @@ class fortran_file:
             return self.scope_list
         scope_list = []
         for scope in self.scope_list:
-            if line_number >= scope.sline and line_number <= scope.eline:
+            if (line_number >= scope.sline) and (line_number <= scope.eline):
                 scope_list.append(scope)
                 for ancestor in scope.get_ancestors():
                     scope_list.append(ancestor)
-        if len(scope_list) == 0 and self.none_scope is not None:
+        if (len(scope_list) == 0) and (self.none_scope is not None):
             return [self.none_scope]
         return scope_list
 
@@ -897,10 +1134,10 @@ class fortran_file:
         curr_scope = None
         for scope in self.scope_list:
             if scope.sline > scope_sline:
-                if line_number >= scope.sline and line_number <= scope.eline:
+                if (line_number >= scope.sline) and (line_number <= scope.eline):
                     curr_scope = scope
                     scope_sline = scope.sline
-        if curr_scope is None and self.none_scope is not None:
+        if (curr_scope is None) and (self.none_scope is not None):
             return self.none_scope
         return curr_scope
 
@@ -945,12 +1182,14 @@ class fortran_file:
             added_entities = include_path[2]
             if file_path in workspace:
                 include_obj = workspace[file_path]["ast"]
-                if (include_obj.none_scope is not None) and (include_obj.none_scope is not parent_scope):
+                if include_obj.none_scope is not None:
+                    if include_obj.inc_scope is None:
+                        include_obj.inc_scope = include_obj.none_scope
                     # Remove old objects
                     for obj in added_entities:
                         parent_scope.children.remove(obj)
                     added_entities = []
-                    for child in include_obj.none_scope.children:
+                    for child in include_obj.inc_scope.children:
                         added_entities.append(child)
                         parent_scope.add_child(child)
                         child.update_fqsn(parent_scope.FQSN)
@@ -958,6 +1197,13 @@ class fortran_file:
                     include_path[2] = added_entities
 
     def close_file(self, line_number):
+        # Close open scopes
+        while self.current_scope is not None:
+            self.end_scope(line_number)
+        # Close and delist none_scope
+        if self.none_scope is not None:
+            self.none_scope.end(line_number)
+            self.scope_list.remove(self.none_scope)
         # Tasks to be done when file parsing is finished
         for private_name in self.private_list:
             obj = self.get_object(private_name)
@@ -967,16 +1213,31 @@ class fortran_file:
             obj = self.get_object(public_name)
             if obj is not None:
                 obj.set_visibility(1)
-        if self.none_scope is not None:
-            self.none_scope.end(line_number)
-            self.scope_list.remove(self.none_scope)
 
     def check_file(self, obj_tree, file_contents):
         errors = []
         tmp_list = self.scope_list
         if self.none_scope is not None:
             tmp_list += [self.none_scope]
+        for error in self.end_errors:
+            errors.append({
+                "range": {
+                    "start": {"line": error[1]-1, "character": 0},
+                    "end": {"line": error[1]-1, "character": 0}
+                },
+                "message": 'Unexpected end of scope at line {0}'.format(error[0]),
+                "severity": 1
+            })
         for scope in tmp_list:
+            if not scope.check_valid_parent():
+                errors.append({
+                    "range": {
+                        "start": {"line": scope.sline-1, "character": 0},
+                        "end": {"line": scope.sline-1, "character": 0}
+                    },
+                    "message": 'Invalid parent for "{0}" declaration'.format(scope.get_desc()),
+                    "severity": 1
+                })
             for error in scope.check_double_def(file_contents, obj_tree):
                 # Check preproc if
                 if self.check_ppif(error[1]):
@@ -997,7 +1258,17 @@ class fortran_file:
                             "end": {"line": error[1], "character": error[3]}
                         },
                         "message": 'Variable "{0}" masks variable in parent scope'.format(error[4]),
-                        "severity": 2
+                        "severity": 2,
+                        "relatedInformation": [{
+                            "location": {
+                                "uri": path_to_uri(error[5]),
+                                "range": {
+                                    "start": {"line": error[6], "character": 0},
+                                    "end": {"line": error[6], "character": 0}
+                                }
+                            },
+                            "message": ""
+                        }]
                     })
             for error in scope.check_use(obj_tree, file_contents):
                 # Check preproc if
@@ -1009,6 +1280,6 @@ class fortran_file:
                         "end": {"line": error[0], "character": error[2]}
                     },
                     "message": 'Module "{0}" not found in project'.format(error[3]),
-                    "severity": 2
+                    "severity": 3
                 })
         return errors
